import jax.numpy as jnp
import jax.random as jr 


def linearized_model(_alpha, fiducial_dv, alpha, derivatives):
    return fiducial_dv + jnp.dot(_alpha - alpha, derivatives)


def simulator(key, parameters, alpha, fiducial_dv, derivatives, covariance):
    d = jr.multivariate_normal(
        key=key, 
        mean=linearized_model(
            _alpha=parameters, 
            fiducial_dv=fiducial_dv, 
            alpha=alpha, 
            derivatives=derivatives
        ),
        cov=covariance
    ) 
    return d


def _mle(d, pi, Finv, mu, dmu, precision):
    return pi + jnp.linalg.multi_dot([Finv, dmu, precision, d - mu])


from typing import List
import equinox as eqx
from tensorflow_probability.substrates.jax.distributions import Distribution

from .ensemble import Ensemble


def default(v, d):
    return v if v is not None else d


class MultiEnsemble(eqx.Module):
    ensembles: List[Ensemble]
    prior: Distribution

    def __init___(self, ensembles, prior):
        self.ensembles = ensembles
        self.prior = prior # Allow to be overwritten in inference call

    def get_multi_ensemble_log_prob_fn(self, datavectors, prior=None):
        _prior = default(prior, self.prior)

        def _multi_ensemble_log_prob_fn(theta):
            # Loop over matched ensembles / datavectors
            L = 0.
            for ensemble, datavector in zip(self.ensembles, datavectors):
                ensemble_log_L = ensemble.ensemble_likelihood(datavector)(theta)
                L = L + ensemble_log_L
            L = L + _prior.log_prob(theta)
            return L 

        return _multi_ensemble_log_prob_fn

    def load_ensembles(self, paths, ensembles):
        # Load sub-ensembles
        self.ensembles = [
            eqx.tree_deserialise_leaves(path, ensemble)
            for path, ensemble in zip(paths, ensembles)
        ]

    def save_ensemble(self, path):
        eqx.tree_serialise_leaves(path, self)

    def load_ensemble(self, path):
        return eqx.tree_deserialise_leaves(path, self)


import os
import numpy as np 
import matplotlib.pyplot as plt
from chainconsumer import Chain, ChainConfig, ChainConsumer, PlotConfig, Truth

from utils import make_df

"""
    Posterior plotting etc
"""

def plot_parameters_and_summaries(parameters, summaries, param_names, filename):
    c = ChainConsumer()
    c.add_chain(
        Chain(
            samples=make_df(parameters, np.ones((len(parameters),)), param_names), 
            name="Params", 
            color="blue", 
            plot_cloud=True, 
            plot_contour=False
        )
    )
    c.add_chain(
        Chain(
            samples=make_df(summaries, np.ones((len(summaries),)), param_names), 
            name="Summaries", 
            color="red", 
            plot_cloud=True, 
            plot_contour=False
        )
    )
    c.add_truth(Truth(location=dict(zip(param_names, alpha)), name=r"$\pi^0$"))
    fig = c.plotter.plot()
    plt.savefig(filename)#os.path.join(results_dir, "params.pdf"))
    plt.close()


def plot_posterior(
    samples,
    samples_log_prob,
    alpha,
    X_,
    Finv,
    alpha_=None,
    Finv_est=None,
    param_names=None,
    filename=None
):
    posterior_df = make_df(samples, samples_log_prob, param_names)

    c = ChainConsumer()

    # Fisher truth
    fisher_chain = Chain.from_covariance(
        alpha,
        Finv,
        columns=param_names,
        name="Fisher",
        color="k"
    )
    c.add_chain(fisher_chain)

    # Fisher estimated
    c.add_chain(
        Chain.from_covariance(
            X_,
            Finv_est if Finv_est is not None else Finv,
            columns=param_names,
            name=r"$F_{\Sigma^{-1}}$",
            color="k",
            linestyle=":"
        )
    )

    # SBI posterior
    c.add_chain(Chain(samples=posterior_df, name="SBI", color="r"))

    # True parameters
    c.add_truth(Truth(location=dict(zip(param_names, alpha)), name=r"$\pi$"))

    # MAP
    if alpha_ is not None:
        c.add_truth(Truth(location=dict(zip(param_names, alpha_)), name=r"$\pi$'"))

    # Summary
    c.add_marker(location=dict(zip(param_names, np.asarray(X_))), name=r"$\hat{x}$", color="b")

    # Truth (fiducial)
    c.add_marker(location=dict(zip(param_names, np.asarray(alpha))), name=r"$\alpha$", color="#7600bc")

    # Truth sampled
    c.add_marker(location=dict(zip(param_names, np.asarray(alpha_))), name=r"$\alpha'$", color="g")

    fig = c.plotter.plot()
    plt.savefig(filename)
    plt.close()


import jax
import jax.random as jr
import jax.numpy as jnp
import equinox as eqx
from jaxtyping import Key, Array
from flowjax.flows import block_neural_autoregressive_flow, masked_autoregressive_flow
from flowjax.distributions import Normal


class MAF(eqx.Module):
    flow: eqx.Module
    base_dist: eqx.Module
    scaler: eqx.Module

    def __init__(
        self, 
        event_dim, 
        context_dim, 
        width_size, 
        n_layers, 
        nn_depth, 
        activation=jax.nn.tanh, 
        scaler=None,
        *, 
        key
    ):
        self.base_dist = Normal(
            loc=jnp.zeros(event_dim), scale=jnp.ones(event_dim)
        )
        self.flow = masked_autoregressive_flow(
            key,
            base_dist=self.base_dist,
            cond_dim=context_dim,
            flow_layers=n_layers,
            nn_width=width_size,
            nn_depth=nn_depth,
            nn_activation=activation
        )
        self.scaler = scaler

    def log_prob(self, x, y, **kwargs):
        if self.scaler is not None:
            x, y = self.scaler.forward(x, y)
        return self.flow.log_prob(x, y)

    def loss(self, x, y, **kwargs):
        return -self.log_prob(x, y, **kwargs)

    def sample_and_log_prob_n(
        self, 
        key: Key, 
        y: Array, 
        n_samples: int
    ):
        samples = self.flow.sample(key, (n_samples,), condition=y)
        log_probs = self.flow.log_prob(samples, y)
        return samples, log_probs


import jax.numpy as jnp
import pandas as pd


def make_df(samples, log_probs, param_names):
    df = pd.DataFrame(samples, columns=param_names).assign(log_posterior=log_probs)
    return df

def nan_to_value(samples, log_probs):
    # Set any bad samples to very low probability
    log_probs = log_probs.at[~jnp.isfinite(log_probs)].set(-1e-100)
    return samples, log_probs


import argparse


def get_n_sims_values():
    return [
        500, 600, 700, 800, 900, 1000, 1200, 1400, 
        1600, 1800, 2000, 3000, 4000, 6000, 8000, 
        10_000#, 20_000 
    ]


def get_sbi_types():
    return "NPE", "NLE"


def get_nde_types():
    return "CNF", "MAF"


def get_args():
    parser = argparse.ArgumentParser(
        description="Run Dodelson experiment for linearised model, either NLE or NPE."
    )
    parser.add_argument(
        "-n", '--n_sims', type=int, help="Number of sims.", default=1000
    )
    parser.add_argument(
        "-e", "--n_ndes", type=int, help="Number of flows.", default=1
    )
    parser.add_argument(
        "-o", "--n_obs", type=int, help="Number of mock-data realisations to sample.", default=1
    )
    parser.add_argument(
        "-m", "--model_type", type=str, help="Type of model density estimator", default="CNF"
    )
    parser.add_argument(
        "-s", "--seed", type=int, help="Seed for random number generation.", default=0
    )
    parser.add_argument(
        "-d", "--exp_name", type=str, help="Directory to save runs in.", default=""
    )
    parser.add_argument(
        "--perfect-score", 
        action=argparse.BooleanOptionalAction, 
        help="Score compressor uses as many n_sims as --n_sims."
    )
    parser.add_argument(
        "--eig-sampling", 
        action=argparse.BooleanOptionalAction, 
        help="Sample simulation parameters parallel to Fisher eigendirections or not."
    )
    args = parser.parse_args()
    return args


def save_posterior_objects(
    samples,
    samples_log_prob,
    alpha__log_prob,
    alpha_,
    X_,
    *,
    results_dir
):
    np.save(
        os.path.join(
            results_dir, "posteriors/alpha_samples.npy"
        ), 
        samples
    )
    np.save(
        os.path.join(
            results_dir, "posteriors/alpha_samples_log_prob.npy"
        ), 
        samples_log_prob
    )
    np.save(
        os.path.join(
            results_dir, "posteriors/alpha__log_prob.npy"
        ), 
        alpha__log_prob
    )
    np.save(
        os.path.join(
            results_dir, "data/alpha_.npy"
        ), 
        alpha_
    )
    np.save(
        os.path.join(
            results_dir, "data/alpha_datavector.npy"
        ), 
        X_ 
    )


from typing import Tuple, Sequence
import jax
import jax.numpy as jnp
import jax.random as jr 
import equinox as eqx
from jaxtyping import Key, Array


def default_weights(weights, ndes):
    return weights if weights is not None else jnp.ones((len(ndes))) / len(ndes)


class Ensemble(eqx.Module):
    sbi_type: str
    ndes: Tuple[eqx.Module]
    weights: Array

    def __init__(
        self, ndes: Sequence[eqx.Module], sbi_type: str = "nle", weights: Array = None
    ):
        self.ndes = ndes
        self.sbi_type = sbi_type
        self.weights = default_weights(weights, ndes)

    def nde_log_prob_fn(self, nde, prior, data):
        """ 
            Get log-probability function for NDE at given observation.
        """
        def _nde_log_prob_fn(theta, **kwargs): 
            return (
                nde.log_prob(x=data, y=theta, **kwargs) 
                + prior.prior.log_prob(theta)
            )
        return _nde_log_prob_fn

    def ensemble_log_prob_fn(self, data, prior=None):
        """ 
            Get log-probability function for NDE at given observation 
            for whole ensemble of NDEs.
            - some NDEs may have a probabilistic estimate of the likelihood
              so a key is provided, the ndes are set to inference mode to 
              imply this key is not used for dropout etc.
        """
        if self.sbi_type == "nle":
            def _joint_log_prob_fn(theta, key=None):
                L = 0.
                for n, (nde, weight) in enumerate(zip(self.ndes, self.weights)):
                    if key is not None:
                        key = jr.fold_in(key, n)
                    nde_log_L = nde.log_prob(x=data, y=theta, key=key)
                    L_nde = weight * jnp.exp(nde_log_L)
                    L = L + L_nde
                L = jnp.log(L) 
                if prior is not None:
                    L = L + prior.log_prob(theta)
                return L

        if self.sbi_type == "npe":
            def _joint_log_prob_fn(theta, key=None):
                L = 0.
                for n, (nde, weight) in enumerate(zip(self.ndes, self.weights)):
                    if key is not None:
                        key = jr.fold_in(key, n)
                    nde_log_L = nde.log_prob(x=theta, y=data, key=key)
                    L_nde = weight * jnp.exp(nde_log_L)
                    L = L + L_nde
                return jnp.log(L) 

        return _joint_log_prob_fn

    def _ensemble_log_prob_fn(self, datavectors, prior=None):
        """ 
            Get log-probability function for NDE at given observation 
            for whole ensemble of NDEs.
            - some NDEs may have a probabilistic estimate of the likelihood
              so a key is provided, the ndes are set to inference mode to 
              imply this key is not used for dropout etc.
        """
        # if datavectors is a list, jax.tree_map(lambda x: nde.log_prob(x=x, y=theta, key=key))
        # > assert isinstance(datavectors, List[Array, ...])
        # if its an array, check if it has a 'batch dim' 
        # 

        if self.sbi_type == "nle":

            # def _joint_log_prob_fn(theta, key=None):
            #     Ls = jax.tree_map(
            #         lambda nde, x: jnp.exp(nde.log_prob(x=x, y=theta, key=key)), self.ndes
            #     )
            #     return Ls

            def _joint_log_prob_fn(theta, key=None):
                L = 0.
                for n, (nde, weight) in enumerate(zip(self.ndes, self.weights)):
                    if key is not None:
                        key = jr.fold_in(key, n)
                    nde_log_L = nde.log_prob(x=data, y=theta, key=key)
                    L_nde = weight * jnp.exp(nde_log_L)
                    L = L + L_nde
                L = jnp.log(L) 
                if prior is not None:
                    L = L + prior.log_prob(theta)
                return L

        if self.sbi_type == "npe":
            def _joint_log_prob_fn(theta, key=None):
                L = 0.
                for n, (nde, weight) in enumerate(zip(self.ndes, self.weights)):
                    if key is not None:
                        key = jr.fold_in(key, n)
                    nde_log_L = nde.log_prob(x=theta, y=data, key=key)
                    L_nde = weight * jnp.exp(nde_log_L)
                    L = L + L_nde
                return jnp.log(L) 

        return _joint_log_prob_fn

    def ensemble_likelihood(self, data):
        return self.ensemble_log_prob_fn(data, prior=None)

    def calculate_stacking_weights(self, losses):
        """
            Calculate weightings of NDEs in ensemble
            - losses is a list of final-epoch validation losses
        """
        Ls = jnp.array(
            [-losses[n] for n, _ in enumerate(self.ndes)]
        )
        nde_weights = jnp.exp(Ls) / jnp.exp(Ls).sum() #jax.nn.softmax(Ls)
        return nde_weights

    def save_ensemble(self, path):
        eqx.tree_serialise_leaves(path, self)

    def load_ensemble(self, path):
        return eqx.tree_deserialise_leaves(path, self)



import warnings
import os
import time
from datetime import datetime
import matplotlib.pyplot as plt
import equinox as eqx
import optuna
from optuna.pruners import BasePruner
import plotly.graph_objects as go


class DuplicateIterationPruner(BasePruner):
    """
    DuplicatePruner

    https://github.com/optuna/optuna/issues/2021#issuecomment-1702539660

    Pruner to detect duplicate trials based on the parameters.

    This pruner is used to identify and prune trials that have the same set of parameters
    as a previously completed trial.
    """

    def prune(
        self, study: optuna.study.Study, trial: optuna.trial.FrozenTrial
    ) -> bool:
        completed_trials = study.get_trials(
            states=[optuna.trial.TrialState.COMPLETE]
        )

        for completed_trial in completed_trials:
            if completed_trial.params == trial.params:
                return True

        print("Pruned duplicate...")

        return False


def get_args():
    parser = argparse.ArgumentParser(
        description="Run architecture search for NPE or NLE."
    )
    parser.add_argument(
        "--exp_name",
        type=str,
        default="",
        help="Name of arch search experiment."
    )
    parser.add_argument(
        "--model_type",
        type=str,
        default="CNF",
        help="Density estimator model type."
    )
    parser.add_argument(
        "--multiprocess", 
        action=argparse.BooleanOptionalAction, 
        help="Run in parallel or not."
    )
    parser.add_argument(
        "--n_processes", 
        type=int,
        default=0,
        help="Run in parallel or not."
    )
    args = parser.parse_args()
    return args


def get_trial_hyperparameters(trial, model_type):
    # Arrange hyperparameters to optimise for and return to the experiment
    if model_type == "CNF":
        model_hyperparameters = {
            "width" : trial.suggest_int(name="width", low=3, high=7, step=1), # NN width
            "depth" : trial.suggest_int(name="depth", low=0, high=2, step=1), # NN depth
            "dt" : trial.suggest_float(name="dt", low=0.01, high=0.15, step=0.01), # ODE solver timestep
            "solver" : trial.suggest_categorical(name="solver", choices=["Euler", "Heun", "Tsit5"]),
            "dropout" : 0.
        }
    if model_type == "MAF":
        model_hyperparameters = {
            "width" : trial.suggest_int(name="width", low=3, high=7, step=1), # Hidden units in NNs
            "depth" : trial.suggest_int(name="depth", low=1, high=5, step=1), # Flow depth
            "layers" : trial.suggest_int(name="layers", low=1, high=3, step=1), # NN layers
            "dropout" : 0.
        }
    training_hyperparameters = {
        # Training
        "bs" : trial.suggest_int(name="bs", low=40, high=100, step=10), 
        "lr": trial.suggest_float(name="lr", low=1e-5, high=1e-3, log=True), 
        "p" : trial.suggest_int(name="p", low=10, high=100, step=10),
        "opt" : "adamw" 
    }
    return {**model_hyperparameters, **training_hyperparameters} 


def callback(study: optuna.Study, trial: optuna.Trial, figs_dir: str, arch_search_dir: str) -> None:
    try:
        print("@" * 80 + datetime.today().strftime('%Y-%m-%d %H:%M:%S'))
        print(f"Best values so far:\n\t{study.best_trial}\n\t{study.best_trial.params}")
        print("@" * 80 + "n_trials=" + str(len(study.trials)))

        layout_kwargs = dict(template="simple_white", title=dict(text=None))
        fig = optuna.visualization.plot_param_importances(study)
        fig.update_layout(**layout_kwargs)
        fig.show()
        fig.write_image(os.path.join(figs_dir, "importances.pdf"))

        fig = optuna.visualization.plot_optimization_history(study)
        fig.update_layout(**layout_kwargs)
        fig.show()
        fig.write_image(os.path.join(figs_dir, "history.pdf"))

        fig = optuna.visualization.plot_contour(study)
        fig.update_layout(**layout_kwargs)
        fig.show()
        fig.write_image(os.path.join(figs_dir, "contour.pdf"))

        fig = optuna.visualization.plot_intermediate_values(study)
        fig.update_layout(**layout_kwargs)
        fig.show()
        fig.write_image(os.path.join(figs_dir, "intermediates.pdf"))

        fig = optuna.visualization.plot_timeline(study)
        fig.update_layout(**layout_kwargs)
        fig.show()
        fig.write_image(os.path.join(figs_dir, "timeline.pdf"))

        df = study.trials_dataframe()
        df.to_pickle(os.path.join(arch_search_dir, "arch_search_df.pkl")) # Where to save it, usually as a .pkl
    except ValueError:
        pass # Not enough trials to plot yet


# @eqx.filter_jit
# def kl_loss(key, model, X, Y, Finv, preprocess_fn, n_samples_per_test):
#     """ KL divergence between posterior samples and true posterior """
#     model = eqx.nn.inference_mode(model, True)

#     # Gaussian posterior pdf at each point in parameter space (NOTE: flow returns log probs!)
#     gaussian = lambda y, x, Finv: jss.multivariate_normal.logpdf(y, mean=x, cov=Finv) # Posterior around MAP
#     model_log_prob = lambda y, x: model.log_prob(y, x)

#     def _kl(x, y):
#         # KL divergence between p and q for one x and many posterior samples y
#         q_pi_x = jax.vmap(gaussian, in_axes=(0, None, None))(y, x, Finv)
#         p_pi_x = jax.vmap(model_log_prob, in_axes=(0, None))(y, x) # NPE
#         # return 0.5 * (
#         #     jnp.log(jnp.exp(p_pi_x) / jnp.exp(q_pi_x)) + jnp.log(jnp.exp(q_pi_x) / jnp.exp(p_pi_x)) 
#         # )
#         # return jnp.log(jnp.exp(p_pi_x) / jnp.exp(q_pi_x)) 
#         return jnp.log(jnp.exp(q_pi_x) / jnp.exp(p_pi_x)) # Same "direction" of KL as training
#         # return 0.5 * jnp.sum(
#         #     jsp.kl_div(jnp.exp(p_pi_x), jnp.exp(q_pi_x)), 
#         #     jsp.kl_div(jnp.exp(q_pi_x), jnp.exp(p_pi_x))
#         # )

#     def _sample_posterior(key, x):
#         # Get posterior samples given datavector
#         samples, _ = sample_nde(
#             model,
#             key, 
#             x, 
#             preprocess_fn=preprocess_fn,
#             n_samples=(n_samples_per_test,)
#         )
#         return samples

#     # Get all posterior samples from all datavectors (n_test_data, n_samples_per_test)
#     keys = jr.split(key, len(X))
#     Y = jax.vmap(_sample_posterior)(keys, X)

#     # Apply KL loss to each posterior set of samples and the datavector used to derive them
#     kl_loss = jax.vmap(_kl)(X, Y).mean() # NOTE: not negative: minimise it.

#     print("kl loss", kl_loss)
#     return kl_loss


# @eqx.filter_jit
# def kl_loss_nle(model, simulations, parameters, Finv, parameter_prior, preprocess_fn):
#     model = eqx.nn.inference_mode(model, True)

#     def posterior_log_prob(x, y):
#         return model.log_prob(x, y) + parameter_prior.log_prob(y)

#     # Log posterior probs given true posterior
#     q_pi_x = jss.multivariate_normal(
#         parameters, mean=simulations, cov=Finv
#     )
#     # Log posterior probs given sbi posterior (no preprocess_fn used here...)
#     p_pi_x = jax.vmap(posterior_log_prob)(simulations, parameters)
#     return -jnp.log(q_pi_x / p_pi_x).mean() # NOTE: negative?



from functools import partial
from typing import Callable, Sequence, Any, Tuple
import jax
import jax.numpy as jnp
import jax.random as jr 
import equinox as eqx
from jaxtyping import Key, Array
from tensorflow_probability.substrates.jax.distributions import Distribution 

Prior = Distribution 


def nde_log_prob(
    nde: eqx.Module, data: Array, theta: Array
) -> Array:
    """ NDE log-likelihood log_prob(data|theta) """
    nde = eqx.nn.inference_mode(nde, True)
    return nde.log_prob(x=data, y=theta) 


def nde_log_prob_posterior(
    nde: eqx.Module, data: Array, theta: Array, prior: Prior
) -> Array:
    """ NDE log-poserior logp(data|theta) + logprior(theta) """
    log_likelihood = nde_log_prob(nde, data, theta) 
    log_prior = prior.log_prob(theta)
    return log_likelihood + log_prior


def get_nde_log_prob_fn(
    nde: eqx.Module, data: Array, prior: Prior
) -> Callable:
    def nde_log_prob_fn(theta):
        return nde_log_prob_posterior(
            nde=nde, 
            data=data, 
            theta=theta, 
            prior=prior
        )
    return eqx.filter_jit(nde_log_prob_fn)


def sample_state(
    key: Key, 
    asymptotic: Any, 
    n_walkers: int, 
    chain_length: int
) -> Array:
    key_sample, key_choice = jr.split(key)
    n_samples = 2 * n_walkers * chain_length

    posterior_samples = asymptotic.sample_n(key_sample, n_samples)
    posterior_weights = jnp.full((n_samples,), 1. / n_samples)

    key, _ = jr.split(key)
    p = posterior_weights / posterior_weights.sum()
    ix = jr.choice(
        key_choice,
        jnp.arange(len(posterior_samples)), 
        p=p,
        replace=False, 
        shape=(2 * n_walkers,)
    )

    state = posterior_samples[ix]
    return state


def _get_ix(
    key: Key, 
    samples: Array, 
    weights: Array, 
    n_walkers: int
) -> Array:
    p = weights / weights.sum()
    ix = jr.choice(
        key,
        jnp.arange(len(samples)), 
        p=p,
        replace=False, 
        shape=(2 * n_walkers,)
    )
    return ix 


def _sample_initial_state(
    key: Key, 
    asymptotic: Any, 
    n_walkers: int, 
    chain_length: int
) -> Tuple[Array, Array]:
    n_samples = 2 * n_walkers * chain_length
    samples = asymptotic.sample_n(key, n_samples)
    weights = jnp.full((n_samples,), 1. / n_samples)
    return samples, weights


def sample_nde_state(
    key: Key, 
    nde: eqx.Module, 
    asymptotic: Any, 
    n_walkers: int, 
    chain_length: int
) -> Array:
    """ 
        Sample posterior samples / weights for state from Fisher results
        or a new walkers state
    """
    if nde.fisher_posterior is not None:
        samples = nde.fisher_posterior.samples 
        weights = nde.fisher_posterior.weights
    else:
        key, _ = jr.split(key)
        samples, weights = _sample_initial_state(
            key, asymptotic, n_walkers, chain_length
        )
    key, _ = jr.split(key)
    ix = _get_ix(key, samples, weights, n_walkers)
    state = samples[ix]
    return state


from typing import Optional, Tuple
import jax
import equinox as eqx
from jaxtyping import Float, Array, jaxtyped
from beartype import beartype as typechecker


def stop_grad(a):
    return jax.lax.stop_gradient(a)


class Scaler(eqx.Module):
    x_dim: int
    q_dim: Optional[int] = None
    mu_x: Array
    std_x: Array
    mu_q: Array
    std_q: Array
    use_scaling: bool

    def __init__(
        self, 
        X=None, 
        Q=None, 
        *,
        x_mu_std=None,
        q_mu_std=None,
        use_scaling=True
    ):
        assert not (X is not None and x_mu_std)
        if X is not None:
            self.x_dim = X.shape[-1]
            self.mu_x = X.mean(axis=0)
            self.std_x = X.std(axis=0)
        if x_mu_std is not None:
            self.mu_x, self.std_x = x_mu_std
            self.x_dim = self.mu_x.size

        assert not (Q is not None and q_mu_std)
        if Q is not None:
            self.q_dim = Q.shape[-1] 
            self.mu_q = Q.mean(axis=0)
            self.std_q = Q.std(axis=0)
        if x_mu_std is not None:
            self.mu_q, self.std_q = q_mu_std
            self.q_dim = self.mu_q.size

        self.use_scaling = use_scaling

    @jaxtyped(typechecker=typechecker)
    def forward(
        self, 
        x: Float[Array, "{self.x_dim}"], 
        q: Optional[Float[Array, "{self.q_dim}"]] = None
    ) -> Tuple[Float[Array, "{self.x_dim}"], Float[Array, "{self.q_dim}"]]: 
        if self.use_scaling:
            x = (x - stop_grad(self.mu_x)) / stop_grad(self.std_x)
            q = (q - stop_grad(self.mu_q)) / stop_grad(self.std_q)
        return x, q


import abc
from typing import NamedTuple
import jax
import jax.numpy as jnp
import jax.random as jr
import equinox as eqx
from jaxtyping import Key, Array


class Sample(NamedTuple):
    x: Array 
    y: Array 


def sort_sample(train_mode, simulations, parameters):
    # Sort simulations and parameters according to NPE or NLE
    _nle = train_mode.lower() == "nle"
    return Sample(
        x=simulations if _nle else parameters,
        y=parameters if _nle else simulations 
    )


class _AbstractDataLoader(metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def __init__(self, data, targets, *, key):
        pass

    def __iter__(self):
        raise RuntimeError("Use `.loop` to iterate over the data loader.")

    @abc.abstractmethod
    def loop(self, batch_size):
        pass


class _InMemoryDataLoader(_AbstractDataLoader):
    def __init__(
        self, 
        simulations: Array, 
        parameters: Array, 
        train_mode: str, 
        *, 
        key: Key
    ): 
        self.simulations = simulations 
        self.parameters = parameters 
        self.train_mode = train_mode.lower()
        self.key = key
        assert self.train_mode.lower() in ["nle", "npe"]

    @property 
    def n_batches(self, batch_size):
        return max(int(self.simulations.shape[0] / batch_size), 1)

    def loop(self, batch_size: int):
        # Loop through dataset, batching, while organising data for NPE or NLE
        dataset_size = self.simulations.shape[0]
        one_batch = batch_size >= dataset_size
        key = self.key
        indices = jnp.arange(dataset_size)
        while True:
            # Yield whole dataset if batch size is larger than dataset size
            if one_batch:
                yield sort_sample(
                    self.train_mode, 
                    self.simulations, 
                    self.parameters
                )
            else:
                key, subkey = jr.split(key)
                perm = jr.permutation(subkey, indices)
                start = 0
                end = batch_size
                while end < dataset_size:
                    batch_perm = perm[start:end]
                    yield sort_sample(
                        self.train_mode, 
                        self.simulations[batch_perm], 
                        self.parameters[batch_perm] 
                    )
                    start = end
                    end = start + batch_size


class DataLoader(eqx.Module):
    arrays: tuple[Array, ...]
    batch_size: int
    key: Key

    def __check_init__(self):
        dataset_size = self.arrays[0].shape[0]
        assert all(array.shape[0] == dataset_size for array in self.arrays)

    def __call__(self, step):
        dataset_size = self.arrays[0].shape[0]
        num_batches = dataset_size // self.batch_size
        epoch = step // num_batches
        key = jr.fold_in(self.key, epoch)
        perm = jr.permutation(key, jnp.arange(dataset_size))
        start = (step % num_batches) * self.batch_size
        slice_size = self.batch_size
        batch_indices = jax.lax.dynamic_slice_in_dim(perm, start, slice_size)
        return tuple(array[batch_indices] for array in self.arrays)


from typing import Callable, Tuple, Optional
from functools import partial
import jax
import jax.numpy as jnp
import jax.random as jr
import equinox as eqx
import diffrax as dfx
from jaxtyping import Array, Float, Key, jaxtyped, PRNGKeyArray
from beartype import beartype as typechecker

TimeArray = Float[Array, ""]


def get_timestep_embedding(timesteps, embedding_dim: int):
    """Build sinusoidal embeddings (from Fairseq)."""
    # Convert scalar timesteps to an array
    assert embedding_dim % 2 == 0
    if jnp.isscalar(timesteps):
        timesteps = jnp.array(timesteps)
    timesteps *= 1000.
    half_dim = embedding_dim // 2
    emb = jnp.log(10_000.) / (half_dim - 1.)
    emb = jnp.exp(jnp.arange(half_dim) * -emb)
    emb = timesteps * emb
    emb = jnp.concatenate([jnp.sin(emb), jnp.cos(emb)], axis=0)
    return emb


def get_time_embedder(embedding_dim):
    return (
        partial(get_timestep_embedding, embedding_dim=embedding_dim)
        if embedding_dim is not None 
        else lambda t: t 
    )


def _log_prob_exact(
    t: float | Float[Array, "1"], 
    y: Float[Array, "y"], 
    args: Tuple[None, Float[Array, "q"], eqx.Module, Key]
) -> Tuple[Float[Array, "y"], Float[Array, "1"]]:
    """ Compute trace directly. Use this for low dimensions. """
    y, _ = y
    _, q, func, key, t_embed = args
    t = jnp.atleast_1d(t)

    # fn = lambda y: func(jnp.concatenate([y, t, q]))  
    fn = lambda y: func(y, t_embed(t), q, key=key)
    f, f_vjp = jax.vjp(fn, y) 

    # Compute trace of Jacobian
    (size,) = y.shape
    (dfdy,) = jax.vmap(f_vjp)(jnp.eye(size))
    logp = jnp.trace(dfdy)
    return f, logp


def _log_prob_approx(
    t: TimeArray, 
    y: Float[Array, "y"], 
    args: Tuple[None, Float[Array, "q"], eqx.Module, Key]
) -> Tuple[Float[Array, "y"], Float[Array, "1"]]:
    """ Approx. trace using Hutchinson's trace estimator. """
    y, _ = y
    z, q, func, key, t_embed = args
    t = jnp.atleast_1d(t)
    
    fn = lambda y: func(y, t_embed(t), q, key=key)
    f, f_vjp = jax.vjp(fn, y) 
    
    # Trace estimator
    (z_dfdy,) = f_vjp(z)
    logp = jnp.sum(z_dfdy * z)
    return f, logp


def _get_solver() -> dfx.AbstractSolver:
    # return dfx.Tsit5()
    return dfx.Heun()


class Linear(eqx.Module):
    weight: Array
    bias: Optional[Array] = None
    use_bias: bool = True

    def __init__(
        self, 
        in_size: int, 
        out_size: int, 
        use_bias: bool = True, 
        *, 
        key: Key
    ):
        lim = jnp.sqrt(1. / (in_size + 1.))
        self.weight = jr.truncated_normal(
            key, shape=(out_size, in_size), lower=-2., upper=2.
        ) * lim
        if use_bias:
            self.bias = jnp.zeros((out_size,))
        self.use_bias = use_bias

    def __call__(self, x: Float[Array, "self.in_size"]) -> Float[Array, "self.out_size"]:
        y = self.weight @ x 
        if self.use_bias:
            y = y + self.bias
        return y



class ResidualNetwork(eqx.Module):
    _in: Linear
    layers: Tuple[Linear]
    dropouts: Tuple[eqx.nn.Dropout]
    _out: Linear
    activation: Callable
    dropout_rate: float
    y_dim: int

    def __init__(
        self, 
        in_size: int, 
        out_size: int, 
        width_size: int, 
        depth: int, 
        y_dim: int, 
        activation: Callable,
        dropout_rate: float = 0.,
        *, 
        key: Key
    ):
        """ Time-embedding may be necessary """
        in_key, *net_keys, out_key = jr.split(key, 2 + depth)
        # NOTE: was in_size + y_dim + 1 but t_emb_dim added to y_dim here
        self._in = Linear(in_size + y_dim, width_size, key=in_key)
        layers = [
            Linear(
                width_size + y_dim, width_size, key=_key
            )
            for _key in net_keys 
        ]
        self._out = Linear(width_size, out_size, key=out_key)
        # NOTE: make this optional
        dropouts = [
            eqx.nn.Dropout(p=dropout_rate) for _ in layers
        ]
        self.layers = tuple(layers)
        self.dropouts = tuple(dropouts)
        self.activation = activation
        self.dropout_rate = dropout_rate
        self.y_dim = y_dim
    
    @jaxtyped(typechecker=typechecker)
    def __call__(
        self, 
        x: Float[Array, "{self.in_size}"], 
        t: TimeArray, 
        y: Float[Array, "{self.y_dim}"],
        *, 
        key: PRNGKeyArray
    ) -> Float[Array, "{self.out_size}"]:
        t = jnp.atleast_1d(t)
        xyt = jnp.concatenate([x, y, t])
        h0 = self.activation(self._in(xyt))
        h = h0
        for l, d in zip(self.layers, self.dropouts):
            # Condition on time at each layer
            hyt = jnp.concatenate([h, y, t])
            h = l(hyt)
            h = d(h, key=key)
            h = self.activation(h)
            h = h0 + h
        o = self._out(h)
        return o


class MLP(eqx.nn.MLP):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def __call__(
        self, 
        x: Array, 
        t: float | Array, 
        y: Array, 
        *, 
        key: Key
    ) -> Array:
        return super().__call__(jnp.concatenate([x, t, y]))


class SquashMLP(eqx.Module):
    layers: list[eqx.nn.Linear]
    norms: list[eqx.nn.LayerNorm]

    def __init__(
        self, 
        in_size: int, 
        width_size: int, 
        depth: int, 
        y_dim: int, 
        dropout_rate: float = 0.,
        *, 
        key: Key
    ):
        keys = jr.split(key, depth + 1)
        layers = []
        norms = []
        if depth == 0:
            layers.append(
                ConcatSquash(
                    in_size=in_size, 
                    out_size=in_size, 
                    y_dim=y_dim, 
                    dropout_rate=dropout_rate, 
                    key=keys[0]
                )
            )
        else:
            layers.append(
                ConcatSquash(
                    in_size=in_size, 
                    out_size=width_size, 
                    y_dim=y_dim, 
                    dropout_rate=dropout_rate,
                    key=keys[0]
                )
            )
            # norms.append()#eqx.nn.LayerNorm((width_size,)))
            for i in range(depth - 1):
                layers.append(
                    ConcatSquash(
                        in_size=width_size, 
                        out_size=width_size, 
                        y_dim=y_dim, 
                        dropout_rate=dropout_rate,
                        key=keys[i + 1]
                    )
                )
                # norms.append()#eqx.nn.LayerNorm((width_size,)))
            layers.append(
                ConcatSquash(
                    in_size=width_size, 
                    out_size=in_size, 
                    y_dim=y_dim, 
                    dropout_rate=dropout_rate,
                    key=keys[-1]
                )
            )
        self.layers = layers
        self.norms = norms 

    def __call__(self, x, t, y, key=None):
        t = jnp.atleast_1d(t)
        # for layer, norm in zip(self.layers[:-1], self.norms):
        for layer in self.layers[:-1]:
            if key is not None:
                key, _ = jr.split(key)
            x = layer(x, t, y, key=key)
            # x = norm(x)
            x = jax.nn.tanh(x)
        if key is not None:
            key, _ = jr.split(key)
        x = self.layers[-1](x, t, y, key=key)
        return x


class ConcatSquash(eqx.Module):
    lin1: Linear
    lin2: Linear
    lin3: Linear
    dropout1: eqx.nn.Dropout
    dropout2: eqx.nn.Dropout

    def __init__(self, *, in_size, out_size, y_dim, dropout_rate, key):
        key1, key2, key3 = jr.split(key, 3)
        self.lin1 = Linear(in_size, out_size, key=key1)
        self.lin2 = Linear(y_dim, out_size, key=key2)
        self.lin3 = Linear(y_dim, out_size, use_bias=False, key=key3)
        self.dropout1 = eqx.nn.Dropout(dropout_rate)
        self.dropout2 = eqx.nn.Dropout(dropout_rate)

    def __call__(self, x, t, y, *, key=None):
        if key is not None:
            keys = jr.split(key)
        else:
            keys = [None, None]
        ty = jnp.concatenate([t, y])
        v = self.dropout1(self.lin1(x) * jax.nn.sigmoid(self.lin2(ty)), key=keys[0])
        u = self.dropout2(self.lin3(ty), key=keys[1])
        return v + u


class CNF(eqx.Module):
    net: eqx.Module
    x_dim: int
    y_dim: int
    dt: float
    t1: float
    exact_log_prob: bool 
    solver: dfx.AbstractSolver
    time_embedder: Callable
    scaler: eqx.Module

    def __init__(
        self,
        event_dim: int,
        context_dim: int, 
        width_size: int,
        depth: int,
        activation: Callable,
        dt: float,  
        t1: float,  
        t_emb_dim: int = None,
        exact_log_prob: bool = True, 
        dropout_rate: float = 0.,
        solver: Optional[dfx.AbstractSolver] = None,
        scaler: eqx.Module = None,
        *,
        key: Key
    ):
        if t_emb_dim is not None:
            y_dim = context_dim + t_emb_dim
        else:
            y_dim = context_dim + 1
  
        self.net = SquashMLP(
            in_size=event_dim,
            width_size=width_size,
            depth=depth,
            y_dim=y_dim,
            dropout_rate=dropout_rate,
            key=key
        )
        self.x_dim = event_dim
        self.y_dim = context_dim 
        self.dt = dt
        self.t1 = t1
        self.exact_log_prob = exact_log_prob
        self.solver = solver
        self.time_embedder = get_time_embedder(t_emb_dim)
        self.scaler = scaler
    
    @jaxtyped(typechecker=typechecker)
    def log_prob(
        self, 
        x: Float[Array, "{self.x_dim}"], 
        y: Float[Array, "{self.y_dim}"], 
        key: Optional[PRNGKeyArray] = None,
        exact_log_prob: Optional[bool] = False 
    ) -> Float[Array, ""]:
        solver = _get_solver() if self.solver is None else self.solver

        if self.scaler is not None:
            x, y = self.scaler.forward(x, y)

        if key is not None:
            key, key_eps = jr.split(key)

        args = (y, self.net, key, self.time_embedder)

        # Train or evaluate with exact or estimated Jacobian 
        if exact_log_prob:
            _use_exact_log_prob = exact_log_prob
        else:
            _use_exact_log_prob = self.exact_log_prob

        term = dfx.ODETerm(
            _log_prob_exact if _use_exact_log_prob else _log_prob_approx
        ) 

        if _use_exact_log_prob:
            eps = None
        else:
            eps = jr.normal(key_eps, (self.x_dim,))

        # Add Hutchinson estimator noise samples
        args = (eps,) + args

        x0 = (x, 0.)
        soln = dfx.diffeqsolve(term, solver, 0., self.t1, self.dt, x0, args)
        (z,), (delta_log_likelihood,) = soln.ys
        log_prob = delta_log_likelihood + self.prior_log_prob(z)
        return log_prob

    @jaxtyped(typechecker=typechecker)
    def sample_and_log_prob(
        self, 
        key: PRNGKeyArray, 
        y: Float[Array, "{self.y_dim}"], 
        exact_log_prob: Optional[bool] = None
    ) -> Tuple[Float[Array, "{self.x_dim}"], Float[Array, ""]]:
        """ Sample many samples given a single condition """
        key_eps, key_z, key_sample = jr.split(key, 3)

        args = (y, self.net, key_sample, self.time_embedder)

        # Sample with exact or estimated Jacobian 
        if exact_log_prob is not None:
            _use_exact_log_prob = exact_log_prob
        else:
            _use_exact_log_prob = self.exact_log_prob

        term = dfx.ODETerm(
            _log_prob_exact if _use_exact_log_prob else _log_prob_approx
        ) 

        if _use_exact_log_prob or self.exact_log_prob:
            eps = None
        else:
            eps = jr.normal(key_eps, (self.x_dim,))

        # Add Hutchinson estimator noise samples
        args = (eps,) + args

        # Latent sample
        z = jr.normal(key_z, (self.x_dim,))

        solver = _get_solver() if self.solver is None else self.solver
        delta_log_likelihood = 0.
        x1 = (z, delta_log_likelihood)
        soln = dfx.diffeqsolve(term, solver, self.t1, 0., -self.dt, x1, args)
        (x,), (delta_log_likelihood,) = soln.ys

        log_prob = delta_log_likelihood + self.prior_log_prob(z)
        return x, log_prob

    def sample_and_log_prob_n(
        self, 
        key: Key, 
        y: Array, 
        n_samples: int, 
        exact_log_prob: Optional[bool] = None
    ) -> Tuple[Array, Array]:
        """ Sample x ~ p(x|y) for a fixed y. """

        if exact_log_prob is not None:
            _use_exact_log_prob = exact_log_prob
        else:
            _use_exact_log_prob = self.exact_log_prob

        _sampler = jax.vmap(
            partial(
                self.sample_and_log_prob, 
                exact_log_prob=_use_exact_log_prob
            ), 
            in_axes=(0, None)
        )
        keys = jr.split(key, n_samples)
        samples, log_probs = _sampler(keys, y)
        return samples, log_probs 

    def prior_log_prob(self, z: Array) -> Array:
        return jax.scipy.stats.multivariate_normal.logpdf(
            z, jnp.zeros(self.x_dim), jnp.eye(self.x_dim)
        )

    def loss(self, x: Array, y: Array, key: Optional[Key] = None) -> Array:
        return -self.log_prob(x, y, key=key)


if __name__ == "__main__":
    from typing import Any
    import jax, jax.numpy as jnp, jax.random as jr 
    import equinox as eqx
    import optax
    from sklearn.datasets import make_moons
    import matplotlib.pyplot as plt

    NeuralDensityEstimator = PyTree = Optimiser = Module = OptState = Any

    key = jr.key(0)
    

    @eqx.filter_jit
    def batch_loss_fn(
        nde: eqx.Module, 
        simulations: Array, 
        parameters: Array,
        key: jr.PRNGKey
    ) -> Array:
        nde = eqx.tree_inference(nde, False)
        keys = jr.split(key, len(simulations))
        loss = jax.vmap(nde.loss)(x=simulations, y=parameters, key=keys).mean()
        return loss


    @eqx.filter_jit
    def make_step(
        nde: eqx.Module, 
        simulations: Array, 
        parameters: Array, 
        opt_state: PyTree,
        opt,
        key: jr.PRNGKey
    ) -> Tuple[eqx.Module, OptState, Array]:
        _fn = eqx.filter_value_and_grad(batch_loss_fn)
        L, grads = _fn(nde, simulations, parameters, key=key)
        updates, opt_state = opt.update(grads, opt_state, nde)
        nde = eqx.apply_updates(nde, updates)
        return nde, opt_state, L 


    # @eqx.filter_jit
    def batch_eval_fn(
        nde: eqx.Module, 
        simulations: Array, 
        parameters: Array,
        key: Key
    ) -> Array:
        nde = eqx.tree_inference(nde, True)
        keys = jr.split(key, len(simulations))
        loss = jax.vmap(nde.loss)(x=simulations, y=parameters, key=keys).mean()
        return loss


    X, Y = make_moons(2000, noise=0.05)
    X, Y = jnp.asarray(X), jnp.asarray(Y)[:, jnp.newaxis]
    X = (X - X.mean()) / X.std()

    cnf = CNF(
        X.shape[-1], 
        Y.shape[-1], 
        width_size=8, 
        depth=1, 
        dt=0.01,
        t1=1.,
        activation=jax.nn.tanh, 
        exact_log_prob=False,
        key=key
    )

    opt = optax.adam(2e-4)
    opt_state = opt.init(eqx.filter(cnf, eqx.is_array))
    
    losses = []
    for i in range(10_000):
        key = jr.fold_in(key, i)
        cnf, opt_state, loss = make_step(cnf, X, Y, opt_state, opt, key)
        losses += [loss]
        print(f"\r{i=}, {loss=:.3f}", end="")
        
        if i % 1000 == 0:
            keys = jr.split(key, 10_000)
            Q = jr.choice(key, jnp.array([0., 1.]), (len(keys),))[:, jnp.newaxis]
            sampler = partial(cnf.sample_and_log_prob, exact_log_prob=True)
            samples, _ = jax.vmap(sampler)(keys, Q)

            plt.figure(dpi=200)
            plt.hist2d(*samples.T, bins=200, cmap="PuOr")
            plt.savefig("samples.png")
            plt.close()

            plt.figure(dpi=200)
            plt.plot(losses)
            plt.savefig("loss.png")
            plt.close()



from .ensemble import Ensemble
from .multi_ensemble import MultiEnsemble
from .cnf import CNF
from .maf import MAF
from .ndes import (
    nde_log_prob,
    nde_log_prob_posterior,
    joint_log_prob_posterior,
    get_nde_log_prob_fn,
    sample_nde_state,
    sample_state
)
from .scaler import Scaler


import blackjax.progress_bar
import jax
import jax.random as jr
import jax.numpy as jnp
import blackjax
from blackjax.util import run_inference_algorithm
from blackjax.progress_bar import progress_bar_scan

# def blackjax_sample(key, log_prob_fn, initial_parameters):
#     alg = blackjax.nuts(
#         logdensity_fn=logdensity_fn,
#         inverse_mass_matrix=jnp.eye(2),
#         step_size=1e-3
#     )

#     initial_position = jnp.ones((2,))

#     _ = run_inference_algorithm(
#         rng_key=key,
#         initial_position=initial_parameters,
#         inference_algorithm=alg,
#         num_steps=100,
#         progress_bar=True,
#     )


# def blackjax_sample(key, log_prob_fn, initial_parameters, n_chains=1, n_samples=10_000):

#     key, init_key, warmup_key, sample_key = jr.split(key, 4)

#     warmup = blackjax.window_adaptation(blackjax.nuts, log_prob_fn)

#     def call_warmup(seed, param):
#         (initial_states, tuned_params), _ = warmup.run(seed, param, 1_000)
#         return initial_states, tuned_params

#     warmup_keys = jr.split(warmup_key, n_chains)
#     initial_states, tuned_params = jax.jit(jax.vmap(call_warmup))(warmup_keys, initial_parameters)

#     def inference_loop_multiple_chains(
#         key, initial_states, tuned_params, log_prob_fn, n_samples, n_chains
#     ):
#         kernel = blackjax.nuts.build_kernel()

#         def _step_fn(key, state, **params):
#             return kernel(key, state, log_prob_fn, **params)

#         def _one_step(states, rng_key):
#             keys = jr.split(rng_key, n_chains)
#             states, infos = jax.vmap(_step_fn)(keys, states, **tuned_params)
#             return states, (states, infos)

#         keys = jr.split(key, n_samples)
#         _, (states, infos) = jax.lax.scan(_one_step, initial_states, keys)

#         return (states, infos)

#     states, infos = inference_loop_multiple_chains(
#         sample_key, initial_states, tuned_params, log_prob_fn, n_samples, n_chains
#     )

#     positions = jnp.concatenate(
#         [states.position[:, n, :] for n in range(n_chains)]
#     )
#     densities = jnp.concatenate(
#         [states.logdensity[:, n] for n in range(n_chains)]
#     )
#     return positions, densities


def blackjax_sample(key, log_prob_fn, prior=None, n_chains=1, n_samples=100_000):

    def init_param_fn(seed):
        # Get initial parameters
        return prior.sample(seed=seed)

    warmup = blackjax.window_adaptation(blackjax.nuts, log_prob_fn)

    # we use 4 chains for sampling
    key, init_key, warmup_key, sample_key = jr.split(key, 4)
    init_keys = jr.split(init_key, n_chains)
    init_params = jax.vmap(init_param_fn)(init_keys)

    @jax.vmap
    def call_warmup(seed, param):
        (initial_states, tuned_params), _ = warmup.run(seed, param, 1000)
        return initial_states, tuned_params

    warmup_keys = jr.split(warmup_key, n_chains)
    initial_states, tuned_params = jax.jit(call_warmup)(warmup_keys, init_params)

    def inference_loop_multiple_chains(
        rng_key, initial_states, tuned_params, log_prob_fn, n_samples, num_chains
    ):
        """
            Does this just step EACH sample once? 
            Need to run this for multiple steps?!
        """
        kernel = blackjax.nuts.build_kernel()

        def step_fn(key, state, **params):
            return kernel(key, state, log_prob_fn, **params)

        def one_step(states, i):
            keys = jr.split(jr.fold_in(rng_key, i), num_chains)
            states, infos = jax.vmap(step_fn)(keys, states, **tuned_params)
            return states, (states, infos)

        _, (states, infos) = jax.lax.scan(
            one_step, initial_states, jnp.arange(n_samples)
        )
        return states, infos


    key, sample_key = jr.split(key)
    states, infos = inference_loop_multiple_chains(
        sample_key, initial_states, tuned_params, log_prob_fn, n_samples, n_chains
    )

    return states.position[:, 0], states.logdensity[:, 0]


from typing import Callable, Sequence, Any, Tuple
import jax
import jax.numpy as jnp
import jax.random as jr 
import equinox as eqx

# from .naming import Ensemble, NeuralDensityEstimator, Array, Prior, Key
Prior = Ensemble = NeuralDensityEstimator = Array = Key = Any

"""
    State sampling could be one function... check if nde is fisher trained
"""


def nde_log_prob(
    nde: NeuralDensityEstimator, 
    data: Array, 
    theta: Array, 
    preprocess_fn: Callable | None = None, 
    **kwargs
) -> Array:
    """ NDE log-likelihood log_prob(data|theta) """
    nde = eqx.tree_inference(nde, True)
    _data, _theta = preprocess_fn((data, theta))
    return nde.model.log_prob(x=_data, y=_theta, **kwargs) 


def nde_log_prob_posterior(
    nde: NeuralDensityEstimator, 
    data: Array, 
    theta: Array, 
    prior: Prior, 
    preprocess_fn: Callable | None = None, 
    **kwargs
) -> Array:
    """ NDE log-poserior logp(data|theta) + logprior(theta) """
    log_likelihood = nde_log_prob(
        nde, data, theta, preprocess_fn, **kwargs
    ) 
    log_prior = prior.prior.log_prob(theta)
    return log_likelihood + log_prior


def joint_log_prob_posterior(
    ndes: Sequence[NeuralDensityEstimator], 
    data: Array, 
    theta: Array, 
    prior: Prior, 
    preprocess_fn: Callable | None = None, 
    **kwargs
) -> Array:
    """ Ensemble log-poserior sum_ndes[logp(data|theta) + logprior(theta)] """
    L = 0.
    for nde in ndes:
        nde_log_L = nde_log_prob(
            nde, data, theta, preprocess_fn, **kwargs
        )
        L_nde = nde.stacking_weight * jnp.exp(nde_log_L)
        L = L + L_nde
    return jnp.log(L) + prior.prior.log_prob(theta)


def get_nde_log_prob_fn(
    nde: NeuralDensityEstimator, 
    data: Array, 
    prior: Prior, 
    preprocess_fn: Callable | None = None, 
) -> Callable:
    def nde_log_prob_fn(theta, **kwargs):
        return nde_log_prob_posterior(
            nde=nde, 
            data=data, 
            theta=theta, 
            prior=prior, 
            preprocess_fn=preprocess_fn, 
            **kwargs
        )
    return eqx.filter_jit(nde_log_prob_fn)


def get_ensemble_log_prob_prior_fn(nde_list):
    """
        Get ensemble as an object to use for inference
        so that loading it is easier...
    """
    pass


def sample_state(
    key: Key, 
    asymptotic: Any, 
    n_walkers: int, 
    chain_length: int
) -> Array:
    key, _ = jr.split(key)
    n_samples = 2 * n_walkers * chain_length
    posterior_samples = asymptotic.sample_n(key, n_samples)
    posterior_weights = jnp.full((n_samples,), 1. / n_samples)
    key, _ = jr.split(key)
    p = posterior_weights / posterior_weights.sum()
    ix = jr.choice(
        key,
        jnp.arange(len(posterior_samples)), 
        p=p,
        replace=False, 
        shape=(2 * n_walkers,)
    )
    state = posterior_samples[ix]
    return state


def _get_ix(
    key: Key, 
    samples: Array, 
    weights: Array, 
    n_walkers: int
) -> Array:
    p = weights / weights.sum()
    ix = jr.choice(
        key,
        jnp.arange(len(samples)), 
        p=p,
        replace=False, 
        shape=(2 * n_walkers,)
    )
    return ix 


def _sample_initial_state(
    key: Key, 
    asymptotic: Any, 
    n_walkers: int, 
    chain_length: int
) -> Tuple[Array, Array]:
    n_samples = 2 * n_walkers * chain_length
    samples = asymptotic.sample_n(key, n_samples)
    weights = jnp.full((n_samples,), 1. / n_samples)
    return samples, weights


def sample_nde_state(
    key: Key, 
    nde: NeuralDensityEstimator, 
    asymptotic: Any, 
    n_walkers: int, 
    chain_length: int
) -> Array:
    """ 
        Sample posterior samples / weights for state from Fisher results
        or a new walkers state
    """
    if nde.fisher_posterior is not None:
        samples = nde.fisher_posterior.samples 
        weights = nde.fisher_posterior.weights
    else:
        key, _ = jr.split(key)
        samples, weights = _sample_initial_state(
            key, asymptotic, n_walkers, chain_length
        )
    key, _ = jr.split(key)
    ix = _get_ix(key, samples, weights, n_walkers)
    state = samples[ix]

    return state


# class NDE(eqx.Module):
#     """ Abstract base class for NDEs """
#     def __init__(self):

#     def log_prob(self):

#     def loss(self):

#     def __repr__(self):
#         return nde.name

# class CNF(NDE):